module base;

import kinds;
import varray;
import SDL;
import util;
import std.string;
import opengl;
import std.stdio;
import glutils;
import system;
import ship;
import SDL_mixer;
import sounds;
import rand;
import std.stream;
import std.cstream;
import recorder;
import filepath;
static import stage1dia;
static import stage2dia;
static import stage3dia;
static import stage4dia;
static import stage5dia;
static import stage6dia;
static import stage7dia;

struct Point
{
  import std.math;
    
  double x, y;

  static Point opCall(double x, double y) {
    Point p;
    p.x = x; p.y = y;
    return p;
  }
  
  bool opIn(Rect r) {
    return r.p.x <= x && x <= r.q.x && r.p.y <= y && y <= r.q.y;
  }

  Point opAdd(Point p) {
    return Point(x+p.x, y+p.y);
  }

  Point opSub(Point p) {
    return Point(x-p.x, y-p.y);
  }

  Point opMul_r(double a) {
    return Point(a*x, a*y);
  }
    
  void opAddAssign(Point p) {
    x += p.x;
    y += p.y;
  }

  void opSubAssign(Point p) {
    x -= p.x;
    y -= p.y;
  }
  
  static double dot(Point p, Point q) {
    return p.x*q.x + p.y*q.y;
  }

  static double det(Point p, Point q) {
    return p.x*q.y - p.y*q.x;
  }
  
  double abs() {
    return sqrt(dot(*this, *this));
  }

  void glVertex() {
    glVertex2d(x, y);
  }
  
  static double angle(Point p, Point q) {
    return dot(p,q)/(p.abs*q.abs);
  }

  static Point from_polar_coord(double abs, double rad) {
    return Point(abs*cos(rad), abs*sin(rad));
  }
  
  const Point zero = {x:0, y:0};
  const Point out_of_window = {x:-100, y:-100};
}

alias Point Vector;
  
struct Rect
{
  Point p, q;

  static Rect opCall(double x, double y, double w, double h) {
    Rect r;
    r.p = Point(x, y); r.q = Point(x+w-1, y+h-1);
    return r;
  }
  
  double x() {
    return p.x;
  }
    
  double y() {
    return q.y;
  }
    
  double w() {
    return q.x - p.x + 1;
  }
    
  double h() {
    return q.y - p.y + 1;
  }

  bool is_intersected(Rect r) {
    return 
      ((p.x <= r.p.x && r.p.x <= q.x) || (r.p.x <= p.x && p.x <= r.q.x)) &&
      ((p.y <= r.p.y && r.p.y <= q.y) || (r.p.y <= p.y && p.y <= r.q.y));
  }

  Point center() {
    return 0.5*(p+q);
  }
  
  void draw() {
    glBegin(GL_LINE_LOOP);
    glVertex2d(p.x, p.y);
    glVertex2d(p.x, q.y);
    glVertex2d(q.x, q.y);
    glVertex2d(q.x, p.y);
    glEnd();
  }

  void filled_draw() {
    glRectd(p.x, p.y, q.x, q.y);
  }
}

abstract class Color {
  static const GLdouble[] white = [1, 1, 1];
  static const GLdouble[] red = [1, 0, 0];
  static const GLdouble[] green = [0, 1, 0];
  static const GLdouble[] yellow = [1, 1, 0];
  static const GLdouble[] black = [0, 0, 0];
  
  static const GLdouble[] wall = green;
  static const GLdouble[] shot = [0.3, 0.3, 1];
  static const GLdouble[] capsule = red;
  static const GLdouble[] bullet = white;
  static const GLdouble[] player = red;
  static const GLdouble[] enemy = [1, 1, 0];
  static const GLdouble[] damaged_enemy = red;
}

GLdouble[] RGB(double r, double g, double b) {
  auto col = new GLdouble[3];
  col[0] = r; col[1] = g; col[2] = b;
  return col;
}

interface Worker
{
  Kind kind();
  SubKind subkind();

  void die();
  void _run(Registry reg);
  void _draw(Registry reg);
  bool dead();
}

interface Movable: Worker
{
  Point pos();
}

abstract class BaseWorker: Worker
{
private:
  Kind kind_;
  SubKind subkind_;
  int count_ = 0;
  bool dead_;
public:
  this(SubKind subkind) {
    kind_ = subkind2kind(subkind);
    subkind_ = subkind;
    dead_ = false;
  }
  
  Kind kind(){ return kind_; }
  SubKind subkind(){ return subkind_; }

  int count(){ return count_; }
  
  void die(){ dead_ = true; }
  void run(Registry){}
  void draw(Registry){}
  void _run(Registry reg){run(reg); count_++;}
  void _draw(Registry reg){draw(reg);}
  bool dead(){ return dead_; }
}

abstract class BaseMovable: BaseWorker, Movable
{
private:
  Point pos_;
  Movable parent_;
  
public:
  this(SubKind subkind, Point pos, Movable parent=null) {
    super(subkind);
    pos_ = pos;
    parent_ = parent;
  }

  void die_if_out_of_window() {
    if (!(pos_ in LARGE_WINDOW)) {
      die();
    }
  }

  Point pos() {
    if (parent_)
      return parent_.pos + pos_;
    else
      return pos_;
  }
  void pos(Point p) {
    pos_ = p;
  }

  void move(Vector v) {
    v += pos_;
    pos_ = v;
  }

  void scroll(Registry reg) {
    move(Scroll.search(reg).v);
  }

  Movable parent(){ return parent_; }
}

class Animation
{
private:
  void function()[] drawers_;
  int clock_;
  bool repeat_;
  int fp_;
  static Animation empty_ = null;
  
public:
  this(void function()[] drawers, int fp=1, bool repeat=true) {
    drawers_ = drawers;
    repeat_ = repeat;
    fp_ = fp;
    clock_ = 0;
  }

  this(void function() drawer) {
    auto drawers = new void function()[1];
    drawers[0] = drawer;
    this(drawers);
  }
  
  void countup() {
    ++clock_;
  }

  void draw() {
    void function() drawer;

    if (repeat_)
      drawer = drawers_[(clock_/fp_) % drawers_.length];
    else if (!end())
      drawer = drawers_[clock_/fp_];
    else 
      drawer = drawers_[drawers_.length-1];
      
    drawer();
  }

  bool end(){
    return !repeat_ && (clock_/fp_ >= drawers_.length);
  }

  static Animation empty() {
    if (empty_ is null) 
      empty_ = new Empty();

    return empty_;
  }
  
  static class Empty: Animation {
    this(){super(&do_nothing);}
    void draw(){}
    bool end(){ return true; }
    static void do_nothing(){}
  }
}

template FixedAnimation(alias fname) {
  protected static Animation anim_;
  static this(){ anim_ = new Animation(&fname.draw); }
}

abstract class Animatable: BaseMovable
{
private:
  Animation anim_;
  bool visible_;
  
public:
  double degree, xdegree=0, ydegree=0;
  float scale;
  int ymirror;
  
  GLdouble[] color;
    
  this(SubKind subkind, Point pos, Animation anim,
       GLdouble[] col = Color.white,
       double deg=0, double sc=1.0, Movable parent=null) {
    super(subkind, pos, parent);
    anim_ = anim;
    color = col;
    degree = deg;
    scale = sc;
    visible_ = true;
    ymirror = 1;
  }
    
  void _draw(Registry reg) {
    if (!visible_)
      return;
    
    glPushMatrix();
    glTranslated(pos.x, pos.y, 0);
    glRotated(degree, 0, 0, 1);
    glRotated(xdegree, 1, 0, 0);
    glRotated(ydegree, 0, 1, 0);
    glScaled(ymirror*scale, scale, 1);
    glColor3da(color);
    anim_.draw();
    glPopMatrix();
    super._draw(reg);
  }

  override void _run(Registry reg) {
    super._run(reg);
    anim_.countup();
  }

  bool visible() { return visible_; }
  void visible(bool v){ visible_ = v; }
  
  Animation animation(){ return anim_; }
}
  
class Registry
{
private:
  VArray!(Worker)[Kind.max+1] workers_;
  VArray!(Worker) added_;
  Game root_;
  
public:  
  this(Game root) {
    root_ = root;
    added_ = new VArray!(Worker)();
    for (Kind k=Kind.min; k<=Kind.max; k++) {
      workers_[k] = new VArray!(Worker)();
    }
  }

  void add(Worker wk) {
    added_.add(wk);
  }
    
  void run() {
    foreach (VArray!(Worker) group; workers_) {
      foreach (Worker wk; group) {
        wk._run(this);
      }
    }

    foreach (VArray!(Worker) group; workers_) {
      group.delete_if(delegate(Worker wk){ return wk.dead; });
    }
      
    foreach (wk; added_) {
      workers_[wk.kind].insert_by(wk, delegate(Worker w){ return cast(int)w.subkind(); });
    }
    added_.clear();
  }
    
  void draw() {
    foreach_reverse (VArray!(Worker) group; workers_) {
      foreach_reverse (Worker wk; group) {
        wk._draw(this);
      }
    }
  }

  int num_workers() {
    int sum = 0;
    foreach (VArray!(Worker) group; workers_) {
      sum += group.length;
    }
    return sum;
  }

  Game root() { return root_; }

  T search(T=Worker)(SubKind sub) {
    foreach (wk; workers_[subkind2kind(sub)]) {
      if (wk.subkind == sub) {
        if (cast(T)wk)
          return cast(T)wk;
        else
          throw new Error("search type mismatch");
      }
    }
    return null;
  }

  VArray!(Worker) workers(Kind kind) {
    return workers_[kind];
  }

  void clear() {
    foreach (VArray!(Worker) group; workers_) {
      group.clear();
    }
    added_.clear();
  }

  void clear(Kind kind) {
    workers_[kind].clear();
  }
}

  
class Timer {
private:
  uint fps_, accuracy_, skip_limit_, one_frame_, delay_, old_, skip_;
  uint count_, count_start_;
  double observed_fps_;
  int observed_skip_;
  
public:
  const int N = 12;
  const int DT = 2;
  const int FPS_COUNT = 10;
    
  this(int fps=60, int accuracy=10, int skip_limit=15) {
    fps_ = fps;
    accuracy_ = accuracy*N;
    skip_limit_ = skip_limit;
    one_frame_ = 1000*N/fps;
    delay_ = accuracy - 2;
    observed_fps_ = fps;
    observed_skip_ = 0;
  }

  private uint get_ticks(){ return SDL_GetTicks()*N; };
    
  void reset() {
    old_ = get_ticks();
    skip_ = 0;
      
    count_ = 0;
    count_start_ = SDL_GetTicks();
  }

  void wait_frame(void delegate() dg) {
    uint now = get_ticks();
    uint nxt = old_ + one_frame_;

    if (nxt > now) {
      dg();
      skip_ = 0;
      wait(nxt);
      old_ = nxt;
    } else if (skip_ > skip_limit_) {
      dg();
      skip_ = 0;
      old_ = get_ticks();
    } else {
      skip_++;
      old_ = nxt;
    }

    count_++;
    if (count_ >= FPS_COUNT) {
      uint n = SDL_GetTicks();
      observed_fps_ = 1000.0*count_/(n - count_start_);
      count_ = 0;
      count_start_ = n;
      observed_skip_ = skip_;
      skip_=0;
    }
  }

  double observed_fps(){ return observed_fps_; }
  int observed_skip(){ return observed_skip_; }
  
  private void wait(uint nxt) {
    while (nxt > get_ticks() + accuracy_)
      SDL_Delay(delay_);

    while (nxt > get_ticks()){}
      
  }
}

class Scroll: BaseWorker
{
public:
  Point v = Point.zero;
  
  this() {
    super(SubKind.SCROLL);
  }

  static Scroll search(Registry reg) {
    return reg.search!(Scroll)(SubKind.SCROLL);
  }
}

interface Wall: Worker
{
  Rect rect();
}

interface Shot: Movable
{
  Rect rect();
  void collide_with_wall(Registry, Wall);
  void collide_with_enemy(Registry, Enemy);
  int power();
}

interface Enemy: Movable
{
  Rect rect();
  void collide_with_shot(Registry, Shot);
  void collide_with_wall(Registry, Wall);
  int score();
  int life();
  bool dead();
}

interface Bullet: Worker
{
  Rect rect();
  void collide_with_wall(Registry, Wall);
  void collide_with_charge_blocker(Registry, ChargeBlocker);
}

interface Capsule: Worker
{
  Rect rect();
  void collide_with_player(Registry, Player);
}

template DefineRect(alias p, int w, int h, bool define_draw=false)
{
  Rect rect(){ return Rect(p.x - w/2, p.y - h/2, w, h); };
  static if (define_draw) {
    void draw(Registry reg) {
      if (reg.root.game_config.draw_bounding_box) {
        glColor3d(0.5, 0.5, 0.5);
        rect().draw();
      }
    }
  }
}

abstract class Game
{
  import pb;
  import std.c.stdio;
private:
  int w_, h_, bpp_, fps_;
  char[] title_;
  Uint8* keystate_ = null;
  Timer timer_;
  
  int play_count_;
  int continue_count_;
  int total_continue_count_;
  int cleared_stage_;
  int hiscore_;
  void function(Registry,Point) magazine_ = &NormalShot.create;
  
  bool load_error_flag_;
  GameConfig game_config_;
  KeyConfig keyconfig_;
  
  
  
protected:
  Registry registry_;
  void delegate() reset_func_;
  RNG rng_;

  int stage_;
  int restart_num_=0;
  int score_;
  int rest_players_;
  int extend_count_;
  int speed_;
  int num_options_;
  int blade_count_;
  int d_blade_count_;
  
  static const Event[][][] stages_ =
  [
   stage0dia.events,
   stage1dia.events,
   stage2dia.events,
   stage3dia.events,
   stage4dia.events,
   stage5dia.events,
   stage6dia.events,
   stage7dia.events,
   ];

  static const void function(Registry,Point)[] magazines_ =
  [
   &NormalShot.create,
   &Blade.create,
   &Reflective.create,
   &FiveWay.create,
   &ShortRange.create,
   ];
  
public:
  this(KeyConfig keyconfig = null,
       GameConfig config=null) {
    title_ = "PB";
    
    w_ = 640; h_ = 480; bpp_ = 0; fps_ = 60;
    keyconfig_ = keyconfig;
    timer_ = new Timer(fps_);
    registry_ = new Registry(this);
    reset_func_ = null;
    game_config_ = config;
    load_error_flag_ = false;
  }

  int w(){return w_;}
  int h(){return h_;}
  int play_count(){return play_count_;}
  int continue_count(){return continue_count_;}
  int total_continue_count(){return total_continue_count_;}
  int score(){return score_;}
  int restart_num(){return restart_num_;}
  int hiscore(){return hiscore_;}
  int rest_players(){return rest_players_;}
  int speed(){ return speed_; }
  int num_options(){ return num_options_;}
  int extend_count(){return extend_count_;}
  int stage(){ return stage_; }
  int blade_count(){ return blade_count_; }
  void extend(){rest_players_++; extend_count_++;}
  bool load_error_flag(){return load_error_flag_;}
  void function(Registry,Point) magazine(){ return magazine_; }
  GameConfig game_config(){ return game_config_; }

  abstract void start();
  
  void new_weapon(void function(Registry,Point) weapon) {
    magazine_ = weapon;
  }

  int magazine_index() {
    foreach (int n, void function(Registry,Point) w;magazines_) {
      if (w == magazine_)
        return n;
    }
    throw new Error("Unknown Magazine");
  }
  
  void blade_count_next() {
    blade_count_ += d_blade_count_;
    if (blade_count_ > 4 || blade_count_ < 0)
      d_blade_count_ *= -1;
  }

  void speed_change(){
    speed_ = (speed_ + 1)%3;
  }

  void add_score(int score){
    score_ += score;
    if (hiscore_ < score_)
      hiscore_ = score_;
  }

  void add_option(){
    if (num_options_ < 2)
      num_options_++;
  }

  RNG rng(){ return rng_; }
  
  void init(){
    Uint32 init_flags;
    Uint32 video_flags;

    init_flags =
      game_config_.sound ? SDL_INIT_EVERYTHING : SDL_INIT_EVERYTHING & ~SDL_INIT_AUDIO;
      
    if (SDL_Init(init_flags) < 0)
      raiseSDLError();

    std.c.stdlib.atexit(&SDL_Quit);
    video_flags = game_config_.fullscreen ? SDL_FULLSCREEN|SDL_OPENGL : SDL_OPENGL;
    
    if (!SDL_SetVideoMode(w_, h_, bpp_, video_flags))
      raiseSDLError();

    enable_sounds(game_config_.sound);

    if (game_config_.sound) {
      Mix_OpenAudio(22050, AUDIO_S16, 2, 2048);
      std.c.stdlib.atexit(&Mix_CloseAudio);
    }
      
    Music.load_musics();
    Wave.load_waves();
    
    SDL_WM_SetCaption(toStringz(title_), toStringz(title_));

    if (game_config_.fullscreen)
      SDL_ShowCursor(SDL_DISABLE);
    
    rng_ = new RNG();
    rng_.init_from_system();
    
    glClearColor(0, 0, 0, 0);
    glViewport(0, 0, w_, h_);

    reset_matrix();

    timer_.reset();
  };
  
  void quit(){ std.c.stdlib.exit(0); }

  void reset_matrix() {
    glLoadIdentity();
    glOrtho(0, w_, h_, 0, -1000.0, 1000.0);
  }
  
  double observed_fps(){ return timer_.observed_fps; }
  int observed_skip(){ return timer_.observed_skip; }

  bool replay_mode();
  bool pause();
  
  void handle_event(SDL_Event ev);
  
  void mainloop() {
    for (;;) {
      SDL_Event ev;
      while (SDL_PollEvent(&ev)) {
        handle_event(ev);
      }
        
      keystate_ = SDL_GetKeyState(null);

      if (!pause())
        registry_.run();
      
      timer_.wait_frame({
        glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
        registry_.draw();
        SDL_GL_SwapBuffers();
      });

      if (reset_func_) {
        reset_func_();
        reset_func_ = null;
      }
    }
  }

  void reserve_start_stage(int n, int r);
  void set_restart_point(int r);
  void next_player();
  void save_playdata();
  Input create_input(KeyConfig config);
  void hold_replaydata();
  void save_replaydata();
  
  void goto_stage() {
    registry_.clear();
    registry_.add(new SystemInfo());
    registry_.add(new GameInfo());
    registry_.add(create_input(keyconfig_));
    Player player = new Player();
    registry_.add(player);
    registry_.add(new ChargeEffect(player));
    registry_.add(new ShiftScreen());
    registry_.add(new DrawWindow());
    registry_.add(new Scroll());
    registry_.add(new ShotWallCollisionDetector());
    registry_.add(new BulletWallCollisionDetector());
    registry_.add(new BulletChargeCollisionDetector());
    registry_.add(new ShotEnemyCollisionDetector());
    registry_.add(new WallEnemyCollisionDetector());
    registry_.add(new PlayerWallCollisionDetector());
    registry_.add(new PlayerEnemyCollisionDetector());
    registry_.add(new PlayerBulletCollisionDetector());
    registry_.add(new PlayerCapsuleCollisionDetector());
    registry_.add(new PlayerReseter());
    registry_.add(new Stage(stages_[stage_][restart_num_]));
    registry_.add(new Extend());
    Music.stages[stage_].play();
    blade_count_ = 0;
    d_blade_count_ = 1;
  }
  
  void show_game_over();
  void show_continue();
  void continue_game();
  void show_title();
  void show_ending();
  
  
  void start_next_game();
  
  void reset_hiscore() {
    hiscore_ = 100;
    play_count_ = 0;
    total_continue_count_ = 0;
    cleared_stage_ = 0;
  }
}

class NormalGame: Game {
private:
  VArray!(Recorder) records_;
  bool pause_;
  
public:
  this(KeyConfig keyconfig = null, GameConfig config=null) {
    super(keyconfig, config);
    load_playdata();
    records_ = new VArray!(Recorder)(15);
    pause_ = false;
  }

  void start() {
    init();
    show_title();
    mainloop();
  }

  void handle_event(SDL_Event ev) {
    switch (ev.type) {
    case SDL_QUIT:
      quit();
    case SDL_KEYDOWN:
      switch (ev.key.keysym.sym) {
      case SDLK_ESCAPE:
        quit();
      case SDLK_p:
        pause_ = !pause;
        break;
      default:
        break;
      }
      break;
    default:
      break;
    }
  }
    
  bool replay_mode(){ return false; }
  bool pause(){ return pause_; }
  
  void goto_stage() {
    save_playdata();
    Game.goto_stage();
    registry_.add(new Recorder(this));
  }
  
  Input create_input(KeyConfig config) {
    return new RealInput(config);
  }
  
  void hold_replaydata() {
    records_.add(cast(Recorder)registry_.search(SubKind.RECORDER));
  }
  
  void save_replaydata() {
    if (!game_config_.replayfile)
      return;

    uint header_size=148+records_.length*8;

    Stream output;

    
    try {
      output = new BufferedFile(game_config_.replayfile,
                                FileMode.Out);
    } catch (OpenException ex) {
      derr.writefln("Cannot write %s", game_config_.replayfile);
      return;
    } finally {
      game_config_.update_replayfilename();
    }

    void write_uint(uint u) {
      for (int i=0; i<4; i++) 
        output.write(cast(ubyte)((u>>i*8)&0xff));
    }
    
    try {
      output.writeString("-PB_REPLAY-\0");
      write_uint(records_.length);
      write_uint(header_size);
      uint size=header_size;
      foreach (Recorder rec; records_) {
        write_uint(size);
        write_uint(rec.datasize);
        size += rec.datasize;
      }
      write_uint(score_);
      write_uint(game_config_.mode);
      write_uint(REPLAY_FILE_VERSION);
      
      for (int i=0; i<116; i++)
        output.write('A');
      
      foreach (Recorder rec; records_)
        rec.save_to_file(output);
    } finally {
      output.close();
    }
  }
  
  void reserve_start_stage(int n, int r) {
    stage_ = n;
    restart_num_ = r;
    reset_func_ = &goto_stage;
  }

  void set_restart_point(int r) {
    restart_num_ = r;
  }

  void next_player() {
    hold_replaydata();
    setup_weapon();
    rest_players_--;
    if (rest_players_ >= 0) {
      reserve_start_stage(stage_, restart_num_);
    } else {
      show_continue();
    }
  }

  void goto_game_over() {
    save_playdata();
    registry_.clear();
    registry_.add(new SystemInfo());
    registry_.add(new RealInput(keyconfig_));
    registry_.add(new ShiftScreen());
    registry_.add(new DrawWindow());
    registry_.add(new GameOver());
  }

  void show_game_over() {
    reset_func_ = &goto_game_over;
  }

  void goto_continue() {
    save_replaydata();
    registry_.clear();
    registry_.add(new SystemInfo());
    registry_.add(new RealInput(keyconfig_));
    registry_.add(new ShiftScreen());
    registry_.add(new DrawWindow());
    registry_.add(new Continue());
  }
  
  void show_continue() {
    reset_func_ = &goto_continue;
  }

  void continue_game() {
    init_next_game(true);
    reserve_start_stage(stage_, restart_num_);
  }

  void init_next_game(bool cont) {
    switch (game_config_.mode) {
    case GameConfig.GameMode.NORMAL:
      score_ = cont ? 1 : 0;
      break;
    case GameConfig.GameMode.TRAINING:
      score_ = cont ? 2 : 3;
      break;
    case GameConfig.GameMode.DEBUG:
      score_ = 4;
      break;
    }
    rest_players_ = game_config_.start_life;
    if (!cont) {
      play_count_++;
      continue_count_ = 0;
      speed_ = 0;
    } else {
      continue_count_++;
      total_continue_count_++;
    }
    extend_count_ = 0;
    setup_weapon();
    rng.init_from_system();
  }

  void start_next_game() {
    init_next_game(false);
    reserve_start_stage(game_config_.start_stage,
                        game_config_.start_point);
  }

  void setup_weapon() {
    num_options_ = game_config_.start_num_options;
    magazine_ = magazines_[game_config_.start_magazine];
  }

  
  void goto_title() {
    registry_.clear();
    registry_.add(new SystemInfo());
    registry_.add(new TextTester());
    registry_.add(new RealInput(keyconfig_));
    registry_.add(new ShiftScreen());
    registry_.add(new DrawWindow());
    registry_.add(new Title());
  }
  
  void show_title() {
    reset_func_ = &goto_title;
  }

  void goto_ending() {
    if (cleared_stage_ < stage_)
      cleared_stage_ = stage_;
    save_playdata();
    hold_replaydata();
    save_replaydata();
    registry_.clear();
    registry_.add(new SystemInfo());
    registry_.add(new RealInput(keyconfig_));
    registry_.add(new ShiftScreen());
    registry_.add(new DrawWindow());
    registry_.add(new GameEnd());
  }

  void show_ending() {
    reset_func_ = &goto_ending;
  }

  void save_playdata() {
    if (cleared_stage_ < stage_ - 1)
      cleared_stage_ = stage_ - 1;
    FILE *output;
    output = fopen(toStringz(score_filename), toStringz("w"));
    if (!output)
      return;
    fprintf(output, std.string.toStringz("%d "), hiscore);
    fprintf(output, std.string.toStringz("%d "), play_count);
    fprintf(output, std.string.toStringz("%d "), total_continue_count);
    fprintf(output, std.string.toStringz("%d \n"), cleared_stage_);
    fprintf(output, std.string.toStringz("hiscore = %d\n"), hiscore);
    fprintf(output, std.string.toStringz("playcount = %d\n"), play_count);
    fprintf(output, std.string.toStringz("continue = %d\n"), total_continue_count);
    fprintf(output, std.string.toStringz("cleared_stage = %d\n"), cleared_stage_);
    fclose(output);
  }

  void load_playdata() {
    Stream input;
    try {
      input = new BufferedFile(score_filename);
    } catch(OpenException ex) {
      load_error();
      return;
    }

    try {
      char[][] words = std.string.split(input.readLine());
      if (words.length != 4) {
        load_error();
        return;
      }

      hiscore_ = atoi(words[0]);
      play_count_ = atoi(words[1]);
      total_continue_count_ = atoi(words[2]);
      cleared_stage_ = atoi(words[3]);
    } finally {
      input.close();
    }
  }

  void load_error() {
    reset_hiscore();
    load_error_flag_ = true;
  }
}

class ReplayGame: Game {
private:
  AllRecords records_;
  int index_;
  bool pause_;
  
  void seek_replay(int idx) {
    pause_ = false;
    index_ = idx;
    if (index_ < 0) {
      index_ = -1;
      reset_func_ = &goto_replay_title;
      return;
    } else if (index_ >= records_.length) {
      index_ = records_.length;
      reset_func_ = &goto_replay_end;
      return;
    }
    reset_func_ = &setup_replay;
  }

  void seek_next() {
    seek_replay(index_+1);
  }

  void seek_prev() {
    seek_replay(index_-1);
  }
  
  void setup_replay() {
    StageRecord rec = records_[index_];
    stage_ = rec.stage;
    restart_num_ = rec.restart_num;
    score_ = rec.score;
    rest_players_ = rec.rest_players;
    extend_count_ = rec.extend_count;
    speed_ = rec.speed;
    num_options_ = rec.num_options;
    new_weapon(magazines_[rec.magazine_index]);
    rng_.init_from_internal_vector(rec.mt, rec.mti);
    goto_stage();
  }

  void goto_replay_title() {
    registry_.clear();
    registry_.add(new SystemInfo());
    registry_.add(new ShiftScreen());
    registry_.add(new DrawWindow());
    registry_.add(new ReplayTitle());
  }

  void goto_replay_end() {
    registry_.clear();
    registry_.add(new SystemInfo());
    registry_.add(new ShiftScreen());
    registry_.add(new DrawWindow());
    registry_.add(new ReplayEnd());
  }
  
public:
  this(KeyConfig keyconfig,
       GameConfig config,
       AllRecords records)
  {
    config.mode = cast(GameConfig.GameMode)records.mode;
    super(keyconfig, config);
    records_ = records;
    index_ = 0;
    pause_ = false;
  }

  bool replay_mode(){ return true; }
  bool pause(){ return pause_; }
  
  void start() {
    init();
    reset_hiscore();
    hiscore_ = records_.score;
    seek_replay(-1);
    mainloop();
  }
  
  void handle_event(SDL_Event ev) {
    switch (ev.type) {
    case SDL_QUIT:
      quit();
    case SDL_KEYDOWN:
      switch (ev.key.keysym.sym) {
      case SDLK_ESCAPE:
        quit();
      case SDLK_RIGHT, SDLK_l:
        seek_next();
        break;
      case SDLK_LEFT, SDLK_h:
        seek_prev();
        break;
      case SDLK_s:
        seek_replay(-1);
        break;
      case SDLK_p:
        pause_ = !pause_;
        break;
      default:
        break;
      }
      break;
    default:
      break;
    }
  }
  
  void reserve_start_stage(int n, int r){ seek_next(); };
  void set_restart_point(int r){}
  void next_player(){ seek_next(); }
  void save_playdata(){};
  Input create_input(KeyConfig config) {
    return new RecordPlayer(records_[index_].record);
  }
  void hold_replaydata(){};
  void save_replaydata(){};

  void show_game_over(){ quit(); }
  void show_continue(){ quit(); }
  void continue_game(){ quit(); }
  void show_title(){ quit(); }
  void show_ending(){ seek_next(); }
  void start_next_game(){ quit(); }
  
}
